<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Resources\ComponentResource;
use App\Models\Component;
use Illuminate\Http\Request;

class ComponentController extends Controller
{
    /**
     * @unauthenticated
     * List all components records and return a collection of components
     * 
     * @param Request $request
     * @return \Illuminate\Http\JsonResponse
     */
    public function list(Request $request) {
        $request->validate([
            // Pagination
            'per_page' => 'nullable|numeric',
            'page' => 'nullable|numeric',

            // Filter
            'type_id' => 'nullable|string|exists:component_types,id',
            'parent_id' => 'nullable|string|exists:components,id',
            'page_id' => 'nullable|string|exists:pages,id',
        ]);

        $component = new Component;
        $type_id = $request->type_id ?? null;
        $page_id = $request->page_id ?? null;
        $parent_id = $request->parent_id ?? null;

        $component = $type_id ? $component->where('type_id', $type_id) : $component;
        $component = $page_id ? $component->where('page_id', $page_id) : $component;
        $component = $parent_id ? $component->where('parent_id', $parent_id) : $component;

        $perPage = $request->per_page ?? 10;
        return response(ComponentResource::collection($component->paginate($perPage)));
    }

    /**
     * Store a newly created resource in storage.
     *
     * @param Request $request
     * @return \Illuminate\Http\JsonResponse
     */
    public function store(Request $request) {
        $request->validate([
            'name' => 'nullable|string',
            'order' => 'nullable|integer',
            'status' => 'in:draft,published',
            'type_id' => 'required|string|exists:component_types,id',
            'page_id' => 'required|string|exists:pages,id',
            'parent_id' => 'nullable|string|exists:components,id',
        ]);

        $component = Component::create($request->all());

        // return ressource with satus code 201
        return response(ComponentResource::make($component), 201);
    }

    /**
     * Display the specified resource.
     *
     * @unauthenticated
     * @param Request $request
     * @param string $id
     * @return \Illuminate\Http\JsonResponse
     */

    public function show(Request $request, $id) {
        $component = Component::findOrFail($id);
        return ComponentResource::make($component);
    }

    /**
     * Update the specified resource in storage.
     *
     * @param Request $request
     * @param string $id
     * @return \Illuminate\Http\JsonResponse
     */

    public function update(Request $request, $id) {
        $request->validate([
            'name' => 'nullable|string',
            'order' => 'nullable|integer',
            'status' => 'in:draft,published',
            'type_id' => 'string|exists:component_types,id',
            'page_id' => 'string|exists:pages,id',
            'parent_id' => 'nullable|string|exists:components,id',
        ]);

        $component = Component::findOrFail($id);
        $component->update($request->all());
        return ComponentResource::make($component);
    }

    /**
     * Remove the specified resource from storage.
     *
     * @param Request $request
     * @param string $id
     * @return \Illuminate\Http\JsonResponse
     */

    public function delete(Request $request, $id) {
        $component = Component::findOrFail($id);
        $component->delete();
        return response()->json(['message' => 'Component deleted successfully']);
    }
}
