<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Resources\RowResource;
use App\Models\Row;
use Illuminate\Http\Request;


/**
 * @author Arleme Johnson
 * Row Controller Class
 */
class RowController extends Controller
{
    /**
     * @unauthenticated
     * List all rows records and return a collection of row resources
     * 
     * @param Request $request
     * @return RowResource::collection
     */
    public function list(Request $request) {
        $request->validate([
            'per_page' => 'nullable|numeric',
            'page' => 'nullable|numeric',
            'section_id' => 'nullable|string|exists:sections,id'
        ]);
        $perPage = $request->per_page ?? 15;
        $section_id = $request->section_id ?? null;

        if ($section_id) {
            $rows = Row::where('section_id', $section_id)->paginate($perPage);
        } else {
            $rows = Row::paginate($perPage);
        }
        return RowResource::collection($rows);
    }

    /**
     * Store a new row record and return the row resource
     * 
     * @param Request $request
     * @return RowResource
     */
    public function store(Request $request) {
        $request->validate([
            'grid' => 'required|in:1,2,3,4',
            'section_id' => 'required|string|exists:sections,id',
            'order' => 'nullable|integer',
            'status' => 'required|in:draft,published'
        ]);

        $row = Row::create($request->all());
        return response(RowResource::make($row), 201);
    }

    /**
     * @unauthenticated
     * Get data of a row record from its id and return the row resource
     * 
     * @param Request $request
     * @param string $id
     * @return RowResource
     * @throws ModelNotFoundException
     */
    public function show(Request $request, $id) {
        $row = Row::findorFail($id);
        return RowResource::make($row);
    }

    /**
     * Update a row record from its id and return the row resource
     * 
     * @param Request $request
     * @param string $id
     * @return RowResource
     * @throws ModelNotFoundException
     */
    public function update(Request $request, $id) {
        $row = Row::findorFail($id);
        $row->update($request->all());
        return RowResource::make($row);
    }

    /**
     * Delete a row record from its id
     * 
     * @param Request $request
     * @param string $id
     * @return JsonResponse
     * @throws ModelNotFoundException
     */
    public function delete(Request $request, $id) {
        $row = Row::findorFail($id);
        $row->delete();
        return response()->json(null, 204);
    }
}
