<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Resources\SectionResource;
use App\Models\Section;
use Illuminate\Http\Request;


/**
 * @author Arleme Johnson
 * Section Controller Class
 */
class SectionController extends Controller
{
    /**
     * @unauthenticated
     * List all sections records and return a collection of section resources
     * 
     * @param Request $request
     * @return SectionResource::collection
     */
    public function list(Request $request) {
        $request->validate([
            'per_page' => 'nullable|numeric',
            'page' => 'nullable|numeric',
            'page_id' => 'nullable|string|exists:pages,id'
        ]);
        $perPage = $request->per_page ?? 10;
        $page_id = $request->page_id ?? null;

        if ($page_id) {
            $sections = Section::where('page_id', $page_id)->paginate($perPage);
        } else {
            $sections = Section::paginate($perPage);
        }
        return SectionResource::collection($sections);
    }

    /**
     * Store a new section record and return the section resource
     * 
     * @param Request $request
     * @return SectionResource
     */
    public function store(Request $request) {
        $request->validate([
            'title' => 'nullable|string',
            'page_id' => 'required|string|exists:pages,id',
            'order' => 'nullable|integer',
            'status' => 'required|in:draft,published'
        ]);

        $section = Section::create($request->all());
        return response(SectionResource::make($section), 201);
    }

    /**
     * @unauthenticated
     * Get data of a section record from its id and return the section resource
     * 
     * @param Request $request
     * @param string $id
     * @return SectionResource
     * @throws ModelNotFoundException
     */
    public function show(Request $request, $id) {
        $section = Section::findorFail($id);
        return SectionResource::make($section);
    }

    /**
     * Update a section record from its id and return the section resource
     * 
     * @param Request $request
     * @param string $id
     * @return SectionResource
     * @throws ModelNotFoundException
     */
    public function update(Request $request, $id) {
        $section = Section::findorFail($id);
        $request->validate([
            'title' => 'required',
            'page_id' => 'required',
            'order' => 'nullable|integer',
            'status' => 'required|in:draft,published'
        ]);

        $section->update($request->all());
        return SectionResource::make($section);
    }

    /**
     * Delete a section record from its id and return a success message
     * 
     * @param Request $request
     * @param string $id
     * @throws ModelNotFoundException
     */
    public function delete(Request $request, $id) {
        $section = Section::findorFail($id);
        $section->delete();
        
        abort(204);
    }
}
